/*
 * Copyright (c) 2018-2020 REV Robotics
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of REV Robotics nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "rev/CANEncoder.h"

#include "rev/CANSparkMaxDriver.h"
#include "rev/CANSparkMax.h"

#include <stdexcept>

using namespace rev;

CANEncoder::CANEncoder(CANSparkMax& device, EncoderType sensorType, int counts_per_rev) 
    : CANSensor(device) {
    // Unhandled CANError is returned by the c_SparkMax_SetSensorType and c_SparkMax_SetCPR calls
    if (!m_encInitialized || m_counts_per_rev != counts_per_rev) {
        m_encInitialized = true;
        m_altEncInitialized = false;
        m_counts_per_rev = counts_per_rev;
        c_SparkMax_SetSensorType(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), static_cast<c_SparkMax_EncoderType>(sensorType));
        // Don't set the CPR if the sensorType is a kHallEffect or cpr = 0
        if (!(sensorType == EncoderType::kHallSensor || m_counts_per_rev == 0)) {
            c_SparkMax_SetCountsPerRevolution(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), m_counts_per_rev);
        }
    }
}

CANEncoder::CANEncoder(CANSparkMax& device, AlternateEncoderType sensorType, int counts_per_rev) 
    : CANSensor(device) {
    if (m_device->m_limitSwitchInitialized) {
        throw std::invalid_argument("Cannot instantiate an alternative encoder while limit switches are enabled");
    }

    m_device->m_altEncInitialized = true;
    // Unhandled CANError is returned by the c_SparkMax_SetSensorType and c_SparkMax_SetCPR calls
    if (!m_altEncInitialized) {
        m_altEncInitialized = true;
        m_encInitialized = false;
    
        c_SparkMax_SetDataPortConfig(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), c_SparkMax_kDataPortAltEncoder);
    }

    if (m_counts_per_rev != counts_per_rev) {
        c_SparkMax_SetAltEncoderCountsPerRevolution(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), counts_per_rev);
        m_counts_per_rev = counts_per_rev;
    }
    
}

CANEncoder::CANEncoder(CANEncoder&& rhs) : CANSensor(std::move(rhs)),
    m_counts_per_rev(std::move(rhs.m_counts_per_rev)),
    m_encInitialized(rhs.m_encInitialized.load()),
    m_altEncInitialized(rhs.m_altEncInitialized.load()) {}

CANEncoder& CANEncoder::operator=(CANEncoder&& rhs) {
  CANSensor::operator=(std::move(rhs));

  m_device = std::move(rhs.m_device);
  m_counts_per_rev = std::move(rhs.m_counts_per_rev);
  m_encInitialized = rhs.m_encInitialized.load();
  m_altEncInitialized = rhs.m_altEncInitialized.load();

  return *this;
}

CANEncoder::CANEncoder(const CANEncoder& rhs) : CANSensor(rhs) {
    m_device = rhs.m_device;
    m_counts_per_rev = rhs.m_counts_per_rev;
    m_encInitialized = rhs.m_encInitialized.load();
    m_altEncInitialized = rhs.m_altEncInitialized.load();
}

double CANEncoder::GetPosition() {
    float tmp;

    if (m_encInitialized) {
        c_SparkMax_GetEncoderPosition(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderPosition(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    }
    return static_cast<double>(tmp);
}

double CANEncoder::GetVelocity() {
    float tmp;
    if (m_encInitialized) {
        c_SparkMax_GetEncoderVelocity(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderVelocity(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    }
    return static_cast<double>(tmp);
}

CANError CANEncoder::SetPosition(double position) {
    c_SparkMax_ErrorCode status;
    if (m_encInitialized) {
        status = c_SparkMax_SetEncoderPosition(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), position);
    } else {
        status = c_SparkMax_SetAltEncoderPosition(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), position);
    }
    return static_cast<CANError>(status);
}

CANError CANEncoder::SetPositionConversionFactor(double factor) {
    c_SparkMax_ErrorCode status;
    if (m_encInitialized) {
        status = c_SparkMax_SetPositionConversionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), factor);
    } else {
        status = c_SparkMax_SetAltEncoderPositionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), factor);
    }
    return static_cast<CANError>(status);
}

CANError CANEncoder::SetVelocityConversionFactor(double factor) {
    c_SparkMax_ErrorCode status;
    if (m_encInitialized) {
        status = c_SparkMax_SetVelocityConversionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), factor);
    } else {
        status = c_SparkMax_SetAltEncoderVelocityFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), factor);
    }
    return static_cast<CANError>(status);
}

double CANEncoder::GetPositionConversionFactor() {
    float tmp;
    if (m_encInitialized) {
        c_SparkMax_GetPositionConversionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderPositionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    }
    return static_cast<double>(tmp);
}

double CANEncoder::GetVelocityConversionFactor() {
    float tmp;
    if (m_encInitialized) {
        c_SparkMax_GetVelocityConversionFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderVelocityFactor(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    }
    return static_cast<double>(tmp);
}

CANError CANEncoder::SetAverageDepth(uint32_t depth) {
    c_SparkMax_ErrorCode status;
    if (m_encInitialized) {
        status = c_SparkMax_SetAverageDepth(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), depth);
    } else {
        status = c_SparkMax_SetAltEncoderAverageDepth(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), depth);        
    }
    return static_cast<CANError>(status);
}

uint32_t CANEncoder::GetAverageDepth() {
    uint32_t tmp;
    if (m_encInitialized) {
        c_SparkMax_GetAverageDepth(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderAverageDepth(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);        
    }
    return static_cast<uint32_t>(tmp);
}

CANError CANEncoder::SetMeasurementPeriod(uint32_t samples) {
    c_SparkMax_ErrorCode status;
    if (m_encInitialized) {
        status = c_SparkMax_SetMeasurementPeriod(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), samples);
    } else {
        status = c_SparkMax_SetAltEncoderMeasurementPeriod(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), samples);        
    }
    return static_cast<CANError>(status);
}

uint32_t CANEncoder::GetMeasurementPeriod() {
    uint32_t tmp;
    if (m_encInitialized) {
        c_SparkMax_GetMeasurementPeriod(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderMeasurementPeriod(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);        
    }
    return static_cast<uint32_t>(tmp);
}

uint32_t CANEncoder::GetCPR() {
    return GetCountsPerRevolution();
}

uint32_t CANEncoder::GetCountsPerRevolution() {
    uint32_t tmp;
    if (m_encInitialized) {
        c_SparkMax_GetCountsPerRevolution(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);
    } else {
        c_SparkMax_GetAltEncoderCountsPerRevolution(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &tmp);        
    }
    return static_cast<uint32_t>(tmp);
}

int CANEncoder::GetID() const {
    if (m_encInitialized) {
        return static_cast<int>((static_cast<c_SparkMax_MotorType>(m_device->GetInitialMotorType()) == c_SparkMax_kBrushless)
            ? FeedbackSensorType::kHallSensor : FeedbackSensorType::kQuadrature);
    } else {
        return static_cast<int>(FeedbackSensorType::kAltQuadrature);
    }
}

CANError CANEncoder::SetInverted(bool inverted)  {
    if (m_encInitialized) {
        if (static_cast<c_SparkMax_MotorType>(m_device->GetInitialMotorType()) == c_SparkMax_kBrushless) {
            throw std::invalid_argument("Not available in Brushless Mode");
        } 

        auto status = c_SparkMax_SetEncoderInverted(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), inverted);
        return static_cast<CANError>(status);
    } else {
        auto status = c_SparkMax_SetAltEncoderInverted(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), inverted);
        return static_cast<CANError>(status);        
    }
}

bool CANEncoder::GetInverted() const {
    uint8_t inverted;
    if (m_encInitialized) {
        c_SparkMax_GetEncoderInverted(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &inverted);
    } else {
        c_SparkMax_GetAltEncoderInverted(static_cast<c_SparkMax_handle>(m_device->m_sparkMax), &inverted);        
    }
    return inverted ? true : false; 
}
